def template1():
    """CoulombMatrix"""
    script = """
                ## (Enter,datasets)
                << host = cheml
                << function = load_xyz_polarizability
                >> coordinates 0

                ## (Represent,molecular descriptors)
                    << host = cheml
                    << function = CoulombMatrix
                    >> 0 molecules
                    >> df 1

                ## (Store,file)
                    << host = cheml
                    << function = SaveFile
                    << filename = CM_features
                    >> 1 df
            """
    return script.strip().split('\n')

def template2():
    """BagofBonds"""
    script = """
                ## (Enter,datasets)
                    << host = cheml
                    << function = load_xyz_polarizability
                    >> coordinates 0

                ## (Represent,molecular descriptors)
                    << host = cheml
                    << function = BagofBonds
                    >> 0 molecules
                    >> df 1

                ## (Store,file)
                    << host = cheml
                    << function = SaveFile
                    << output_directory = descriptors
                    << filename = BoB_features
                    >> 1 df
            """
    return script.strip().split('\n')

def template3():
    """RDKitFingerprint"""
    script = """
                ## (Enter,datasets)
                    << host = cheml
                    << function = load_cep_homo
                    >> smiles 2

                ## (Represent,molecular descriptors)
                    << host = cheml
                    << function = RDKitFingerprint
                    << molfile = @molfile
                    >> 0 molfile
                    >> df 1

                ## (Store,file)
                    << host = cheml
                    << function = SaveFile
                    << format = smi
                    << header = False
                    << filename = smiles
                    >> filepath 0
                    >> 2 df

                ## (Store,file)
                    << host = cheml
                    << function = SaveFile
                    << filename = Fingerprints
                    >> 1 df
            """
    return script.strip().split('\n')

def template4():
    """Dragon"""
    script = """
                ## (Enter,datasets)
                    << host = cheml
                    << function = load_cep_homo
                    >> smiles 0

                ## (Store,file)
                    << host = cheml
                    << function = SaveFile
                    << format = smi
                    << header = False
                    << filename = smiles
                    >> 0 df
                    >> filepath 1

                ## (Store,file)
                    << host = cheml
                    << function = SaveFile
                    << output_directory = Dragon
                    << filename = Dragon_features
                    >> 2 df

                ## (Represent,molecular descriptors)
                    << host = cheml
                    << function = Dragon
                    << molFile = @molfile
                    >> 1 molfile
                    >> df 2
            """
    return script.strip().split('\n')

def template5():
    """composition"""
    script = """
                ## (Enter,datasets)
                    << host = cheml
                    << function = load_comp_energy
                    >> entries 0
                    >> entries 1
                    >> entries 2
                    >> formation_energy 7

                ## (Represent,inorganic descriptors)
                    << host = cheml
                    << function = ElementFractionAttributeGenerator
                    >> 0 entries
                    >> df 5

                ## (Represent,inorganic descriptors)
                    << host = cheml
                    << function = IonicityAttributeGenerator
                    >> 1 entries
                    >> df 3

                ## (Represent,inorganic descriptors)
                    << host = cheml
                    << function = MeredigAttributeGenerator
                    >> 2 entries
                    >> df 4

                ## (Prepare,data manipulation)
                    << host = pandas
                    << function = concat
                    << axis = 1
                    >> 3 df2
                    >> 4 df3
                    >> 5 df1
                    >> df 6

                ## (Enter,python script)
                    << host = cheml
                    << function = PyScript
                    << line01 = print 'features.shape:', iv1.shape
                    << line02 = print 'energies.shape:', iv2.shape
                    >> 6 iv1
                    >> 7 iv2

            """
    return script.strip().split('\n')

def template6():
    """crystal"""
    script = """
                ## (Enter,datasets)
                    << host = cheml
                    << function = load_crystal_structures
                    >> entries 0
                    >> entries 1
                    >> entries 5

                ## (Represent,inorganic descriptors)
                    << host = cheml
                    << function = CoordinationNumberAttributeGenerator
                    >> 0 entries
                    >> df 2

                ## (Represent,inorganic descriptors)
                    << host = cheml
                    << function = CoulombMatrixAttributeGenerator
                    >> 1 entries
                    >> df 3

                ## (Prepare,data manipulation)
                    << host = pandas
                    << function = concat
                    << axis = 1
                    >> 2 df2
                    >> 3 df3
                    >> df 4
                    >> 6 df1

                ## (Enter,python script)
                    << host = cheml
                    << function = PyScript
                    << line01 = print 'shape of features:', iv1.shape
                    >> 4 iv1

                ## (Represent,inorganic descriptors)
                    << host = cheml
                    << function = EffectiveCoordinationNumberAttributeGenerator
                    >> 5 entries
                    >> df 6
            """
    return script.strip().split('\n')

def template7():
    """load_cep_homo"""
    script = """
                ## (Enter,datasets)
                    << host = cheml
                    << function = load_cep_homo
                    >> homo 0
                    >> smiles 3

                ## (Visualize,plot)
                    << host = cheml
                    << function = hist
                    << color = green
                    << x = 0
                    << bins = 20
                    >> 0 dfx
                    >> fig 1

                ## (Store,figure)
                    << host = cheml
                    << function = SavePlot
                    << kwargs = {'normed':True}
                    << output_directory = plots
                    << filename = homo_histogram
                    >> 2 fig

                ## (Visualize,artist)
                    << host = cheml
                    << function = decorator
                    << title = CEP_HOMO
                    << grid_color = b
                    << xlabel = eHOMO (eV)
                    << ylabel = %
                    << grid = True
                    >> 1 fig
                    >> fig 2

                ## (Enter,python script)
                    << host = cheml
                    << function = PyScript
                    << line01 = print iv1.head()
                    >> 3 iv1
             """
    return script.strip().split('\n')

def template8():
    """load_cep_homo"""
    script = """
                ## (Store,figure)
                    << host = cheml
                    << function = SavePlot
                    << kwargs = {'normed':True}
                    << output_directory = plots
                    << filename = amwVSdensity
                    >> 0 fig

                ## (Visualize,artist)
                    << host = cheml
                    << function = decorator
                    << title = AMW vs. Density
                    << grid_color = g
                    << xlabel = density (Kg/m3)
                    << ylabel = atomic molecular weight
                    << grid = True
                    << size = 18
                    >> fig 0
                    >> 4 fig

                ## (Enter,python script)
                    << host = cheml
                    << function = PyScript
                    << line01 = print iv1.head()
                    >> 1 iv1

                ## (Enter,datasets)
                    << host = cheml
                    << function = load_organic_density
                    >> smiles 1
                    >> density 2
                    >> features 3

                ## (Visualize,plot)
                    << host = cheml
                    << function = scatter2D
                    << y = 0
                    << marker = o
                    << x = 'AMW'
                    >> 2 dfy
                    >> 3 dfx
                    >> fig 4
            """
    return script.strip().split('\n')

def template9():
    """load_organic_density"""
    script = """
                ## (Enter,datasets)
                    << host = cheml
                    << function = load_xyz_polarizability
                    >> polarizability 0
                    >> coordinates 1

                ## (Visualize,plot)
                    << host = pandas
                    << function = plot
                    << kind = area
                    < subplots = True
                    >> 0 df
                    >> fig 3

                ## (Enter,python script)
                    << host = cheml
                    << function = PyScript
                    << line01 = print iv1[1]
                    >> 1 iv1

                ## (Store,figure)
                    << host = cheml
                    << function = SavePlot
                    << output_directory = plots
                    << filename = pol_barplot
                    >> 2 fig

                ## (Visualize,artist)
                    << host = cheml
                    << function = decorator
                    << title = polarizability area plot
                    << ylim = (60, None)
                    << grid_linestyle = :
                    << grid = True
                    >> fig 2
                    >> 3 fig
            """
    return script.strip().split('\n')


def template11():
    """Model selection"""
    script ="""
                ## (Enter,datasets)
                    << host = cheml
                    << function = load_xyz_polarizability
                    >> coordinates 0
                    >> polarizability 2

                ## (Represent,molecular descriptors)
                    << host = cheml
                    << function = BagofBonds
                    >> 0 molecules
                    >> df 1

                ## (Prepare,split)
                    << host = sklearn
                    << function = train_test_split
                    << test_size = 0.1
                    << random_state = 0
                    >> 1 dfx
                    >> 2 dfy
                    >> dfx_train 3
                    >> dfy_train 4

                ## (Prepare,scaling)
                    << host = sklearn
                    << function = StandardScaler
                    << func_method = fit_transform
                    >> 3 df
                    >> df 7

                ## (Prepare,scaling)
                    << host = sklearn
                    << function = StandardScaler
                    << func_method = fit_transform
                    >> 4 df
                    >> df 5

                ## (Search,evaluate)
                    << host = sklearn
                    << function = scorer_regression
                    << greater_is_better = False
                    >> scorer 8

                ## (Search,grid)
                    << host = sklearn
                    << function = GridSearchCV
                    << scoring = @scorer
                    << estimator = @estimator
                    << param_grid = {'alpha':[1,0.3,0.1,0.03,0.01]}
                    << cv = @cv
                    >> 5 dfy
                    >> 6 cv
                    >> 7 dfx
                    >> 8 scorer
                    >> cv_results_ 9
                    >> cv_results_ 10
                    >> 11 estimator

                ## (Enter,python script)
                    << host = cheml
                    << function = PyScript
                    << line01 = print iv1.head(10)
                    >> 9 iv1

                ## (Store,file)
                    << host = cheml
                    << function = SaveFile
                    << format = txt
                    << output_directory = gridsearch
                    << header = True
                    << filename = GridSearchCV_results
                    >> 10 df

                ## (Prepare,split)
                    << host = sklearn
                    << function = KFold
                    >> api 6

                ## (Model,regression)
                    << host = sklearn
                    << function = MLPRegressor
                    >> api 11

            """
    return script.strip().split('\n')

def template12():
    """Model selection"""
    script = """
                ## (Enter,datasets)
                    << host = cheml
                    << function = load_xyz_polarizability
                    >> coordinates 0
                    >> polarizability 2

                ## (Represent,molecular descriptors)
                    << host = cheml
                    << function = BagofBonds
                    >> 0 molecules
                    >> df 1

                ## (Prepare,split)
                    << host = sklearn
                    << function = train_test_split
                    << test_size = 0.2
                    << random_state = 0
                    >> 1 dfx
                    >> 2 dfy
                    >> dfx_train 3
                    >> dfy_train 4
                    >> dfx_test 10
                    >> dfy_test 13
                    >> dfy_test 21

                ## (Prepare,scaling)
                    << host = sklearn
                    << function = StandardScaler
                    << func_method = fit_transform
                    >> 3 df
                    >> df 6
                    >> api 9

                ## (Prepare,scaling)
                    << host = sklearn
                    << function = StandardScaler
                    << func_method = fit_transform
                    >> 4 df
                    >> df 5
                    >> api 11

                ## (Search,evaluate)
                    << host = sklearn
                    << function = scorer_regression
                    << greater_is_better = False
                    >> scorer 7

                ## (Model,regression)
                    << host = sklearn
                    << function = KernelRidge
                    << kernel = rbf
                    >> api 15

                ## (Search,grid)
                    << host = sklearn
                    << function = GridSearchCV
                    << scoring = @scorer
                    << param_grid = {'alpha':[1,.3,.1,.03,.01,0.003]}
                    << cv = @cv
                    >> 5 dfy
                    >> 6 dfx
                    >> 7 scorer
                    >> cv_results_ 8
                    >> 15 estimator
                    >> best_estimator_ 17
                    >> 19 cv

                ## (Model,regression)
                    << host = sklearn
                    << function = KernelRidge
                    << func_method = predict
                    >> 16 dfx
                    >> 17 api
                    >> dfy_predict 18

                ## (Store,file)
                    << host = cheml
                    << function = SaveFile
                    << format = txt
                    << output_directory = gridsearch
                    << header = True
                    << filename = GridSearchCV_results
                    >> 8 df

                ## (Prepare,scaling)
                    << host = sklearn
                    << function = StandardScaler
                    << func_method = transform
                    >> 9 api
                    >> 10 df
                    >> df 16

                ## (Search,evaluate)
                    << host = sklearn
                    << function = evaluate_regression
                    << r2_score = True
                    << mean_absolute_error = True
                    << median_absolute_error = True
                    << mean_squared_error = True
                    << root_mean_squared_error = True
                    >> 12 dfy_predict
                    >> 13 dfy
                    >> evaluation_results_ 14

                ## (Prepare,scaling)
                    << host = sklearn
                    << function = StandardScaler
                    << func_method = inverse_transform
                    >> 11 api
                    >> df 12
                    >> 18 df
                    >> df 20

                ## (Store,file)
                    << host = cheml
                    << function = SaveFile
                    << output_directory = results
                    << filename = test_evaluation
                    >> 14 df

                ## (Prepare,split)
                    << host = sklearn
                    << function = LeaveOneOut
                    >> api 19

                ## (Visualize,plot)
                    << host = cheml
                    << function = scatter2D
                    << y = 0
                    << x = 0
                    >> 20 dfx
                    >> 21 dfy
                    >> fig 22

                ## (Visualize,artist)
                    << host = cheml
                    << function = decorator
                    << title = calculated vs. predicted
                    << grid_color = g
                    << xlabel = predicted polarizability (Bohr3)
                    << grid_linestyle = :
                    << ylabel = calculated polarizability (Bohr3)
                    >> 22 fig
                    >> fig 23

                ## (Store,figure)
                    << host = cheml
                    << function = SavePlot
                    << output_directory = results
                    << filename = test_predVScalc
                    >> 24 fig

                ## (Enter,python script)
                    << host = cheml
                    << function = PyScript
                    << line01 = ax = iv1.axes[0]
                    << line02 = ax.plot([90,120],[90,120],'r-')
                    << line03 = ov1 = ax.figure
                    >> 23 iv1
                    >> ov1 24
            """
    return script.strip().split('\n')



import types
from math import sqrt, log
import numpy as np
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.GCLPCalculator import GCLPCalculator

class GCLPAttributeGenerator:
    """Class to compute features based on the T=0K ground state.

    Attributes
    ----------
    GCLPCalculator : GCLPCalculator
        A GCLPCalculator instance.
    count_phases : bool
        Flag to include or exclude the number of phases at equilibrium.

    Notes
    -----
    Features:
    1. Formation energy.
    2. Number of phases in equilibrium.
    3. Distance from closest composition (i.e., ||x_i - x_{i,f}||_2 for each
    component i for phase f).
    4. Average distance from all neighbors.
    5. Quasi-entropy (sum x_i * ln(x_i) where x_i is fraction of phase).
    Certain values of the number of phases in equilibrium and "quasi-entropy"
    are only accessible to systems with larger number of elements. Useful if
    you do not want to consider the number of components in an alloy as a
    predictive variable.

    """

    # Tool used to compute ground states.
    GCLPCalculator = None

    # Whether to include the number of phases at equilibrium.
    count_phases = True

    def set_phases(self, phases, energies):
        """Function to define phases used when computing ground states.

        Parameters
        ----------
        phases : array-like
            Compositions to consider. A list of CompositionEntry's.
        energies : array-like
            Corresponding energies. A list of float values.

        """

        self.GCLPCalculator = GCLPCalculator()
        self.GCLPCalculator.add_phases(phases, energies)

    def set_count_phases(self, count_phases):
        """Function to set variable to count number of phases at equilibrium.
        In some cases, you may want to exclude this as a feature because it is
        tied to the number of components in the compound.

        Parameters
        ----------
        count_phases : bool
            Desired setting.

        """

        self.count_phases = count_phases

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CompositionEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        # Check if the GCLP calculation has been defined.
        if not self.GCLPCalculator:
            raise ValueError("GCLP calculator has not been setup.")

        feat_headers.append("T0K:Enthalpy")
        if self.count_phases:
            feat_headers.append("T0K:NPhasesEquilibrium")
        feat_headers.append("T0K:ClosestPhaseDistance")
        feat_headers.append("T0K:MeanPhaseDistance")
        if self.count_phases:
            feat_headers.append("T0K:QuasiEntropy")

        for entry in entries:
            tmp_list = []

            # Run GCLP.
            l,r = self.GCLPCalculator.run_GCLP(entry)

            # Compute formation energy.
            tmp_list.append(l)

            # Compute number of phases.
            if self.count_phases:
                tmp_list.append(len(r))

            # Compute distances.
            phase_distances = []
            elements = entry.get_element_ids()
            fractions = entry.get_element_fractions()

            for phase in r:
                dist = 0.0
                for i,elem in enumerate(elements):
                    diff = phase.get_element_fraction(id=elem) - fractions[i]
                    dist += diff * diff
                phase_distances.append(sqrt(dist))

            tmp_list.append(min(phase_distances))
            tmp_list.append(np.mean(phase_distances))

            # Compute quasi-entropy.
            if self.count_phases:
                entropy = 0.0
                for f in r.values():
                    entropy += f * log(f)
                tmp_list.append(entropy)

            feat_values.append(tmp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import types
import pandas as pd
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry

class CoordinationNumberAttributeGenerator:
    """Class to compute attributes based on the coordination number. Uses the
    Voronoi tessellation to define the coordination network.

    DEV NOTE (LW 15Jul15): Could benefit from adding a face size cutoff, where
    atoms are only defined as coordinated if the face between them is larger
    than a certain fraction of the surface area of both cells. Otherwise
    faces on the cells that are only present to numerical issues will be
    counted as neighbors. Metallic glass community commonly removes any faces
    smaller than 1% of the total surface area of a cell.
    """

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Crystal structures for which features are to be generated. A list
            of CrystalStructureEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CrystalStructureEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CrystalStructureEntry's.


        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")
        elif (entries and not isinstance(entries[0], CrystalStructureEntry)):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")

        feat_headers.append("mean_Coordination")
        feat_headers.append("var_Coordination")
        feat_headers.append("min_Coordination")
        feat_headers.append("max_Coordination")

        for entry in entries:
            temp_list = []

            output = entry.compute_voronoi_tessellation()

            mean = output.face_count_average()
            variance = output.face_count_variance()
            minimum = output.face_count_minimum()
            maximum = output.face_count_maximum()

            temp_list.append(mean)
            temp_list.append(variance)
            temp_list.append(minimum)
            temp_list.append(maximum)

            feat_values.append(temp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import numpy as np
from numpy.linalg import norm
from .Line import Line

class Plane:
    """Class to represent planes in a three dimensional space.

    Documentation obtained from: http://commons.apache.org/proper/commons-math
    /apidocs/org/apache/commons/math4/geometry/euclidean/threed/Plane.html

    Attributes
    ----------
    u : array-like
        First vector of the plane frame (in plane).
    v : array-like
        Second vector of the plane frame (in plane).
    w : array-like
        Third vector of the plane frame (plane normal).
    origin : array-like
        Origin of the plane frame.
    origin_offset : float
        Offset of the origin with respect to the plane.
    tolerance : float
        Tolerance below which points are considered identical.
    """

    def __init__(self, normal=None, tolerance=None, p=None, plane=None,
                 p1=None, p2=None, p3=None):
        """Function to build a plane normal to a given direction and containing
        the origin.

        If p is specified, the plane contains the point. If plane is
        specified, makes a copy of the plane.

        Parameters
        ----------
        normal : array-like
            Normal direction to the plane.
        tolerance : float
            Tolerance below which points are considered identical.
        p : array-like
            Point belonging to the plane.
        plane : Plane
            Plane to copy.
        p1 : array-like
            Point belonging to the plane.
        p2 : array-like
            Point belonging to the plane.
        p3 : array-like
            Point belonging to the plane.

        Raises
        ------
        Exception
            If norm is zero.
        """

        if plane is None and normal is not None and tolerance is not None:
            n = norm(normal)
            if n < 1e-10:
                raise Exception("Norm is zero!")

            # Third vector of the plane frame (plane normal).
            self.w = normal / n

            # Tolerance below which points are considered identical.
            self.tolerance = tolerance

            #  Offset of the origin with respect to the plane.
            self.origin_offset = -np.dot(p, self.w) if p is not None else 0

            # Origin of the plane frame.
            self.origin = -self.origin_offset * self.w

            # First vector of the plane frame (in plane).
            self.u = self.orthogonal(self.w)

            # Second vector of the plane frame (in plane).
            self.v = np.cross(self.w, self.u)
        elif plane is not None:
            #  Offset of the origin with respect to the plane.
            self.origin_offset = plane.origin_offset

            # Origin of the plane frame.
            self.origin = plane.origin

            # First vector of the plane frame (in plane).
            self.u = plane.u

            # Second vector of the plane frame (in plane).
            self.v = plane.v

            # Third vector of the plane frame (plane normal).
            self.w = plane.w

            # Tolerance below which points are considered identical.
            self.tolerance = plane.tolerance
        elif p1 is not None and p2 is not None and p3 is not None and \
                tolerance is not None:
            v1 = np.array(p1, dtype=float)
            v2 = np.array(p2, dtype=float)
            v3 = np.array(p3, dtype=float)
            n_vec = np.cross(v2 - v1, v3 - v1)
            self.__init__(p=v1, normal=n_vec, tolerance=tolerance)

    def orthogonal(self, w):
        """Function to compute a vector orthogonal to a given vector.

        Documentation obtained from:
        http://commons.apache.org/proper/commons-math/javadocs/api-3.3/org
        /apache/commons/math3/geometry/euclidean/threed/Vector3D.html

        There are an infinite number of normalized vectors orthogonal to the
        instance. This method picks up one of them almost arbitrarily. It is
        useful when one needs to compute a reference frame with one of the
        axes in a predefined direction. The following example shows how to
        build a frame having the k axis aligned with the known vector u :
        Vector3D k = u.normalize();
        Vector3D i = k.orthogonal();
        Vector3D j = Vector3D.crossProduct(k, i);

        Parameters
        ----------
        w : array-like
            Given vector.

        Returns
        -------
        output : array-like
            Normalized orthogonal vector.

        """

        threshold = 0.6 * norm(w)
        if threshold == 0:
            raise Exception("Norm is zero!")
        x = w[0]
        y = w[1]
        z = w[2]
        if abs(x) <= threshold:
            inverse = 1 / np.math.sqrt(y ** 2 + z ** 2)
            return np.array([0, inverse * z, -inverse * y])
        elif abs(y) <= threshold:
            inverse = 1 / np.math.sqrt(x ** 2 + z ** 2)
            return np.array([-inverse * z, 0, inverse * x])
        inverse = 1 / np.math.sqrt(x ** 2 + y ** 2)
        return np.array([inverse * y, -inverse * x, 0])

    def get_normal(self):
        """Function to get the direction normal to the plane.

        Returns
        -------
        output : array-like
            Direction normal to the plane.
        """
        return self.w

    def get_origin(self):
        """Function to get the origin of the plane frame.

        Returns
        -------
        output : array-like
            Origin of the plane frame.
        """
        return self.origin

    def project(self, p):
        """Function to transform a point into a projection.

        Parameters
        ----------
        p : array-like
            Desired point to transform.

        Returns
        -------
        output : array-like
            Transformation.

        """
        p2 = np.array([np.dot(p, self.u), np.dot(p, self.v)])
        return self.u * p2[0] + self.v * p2[1] - self.origin_offset * self.w

    def get_point_at(self, in_plane, offset):
        """Function to get one point from 3D-space.

        Parameters
        ----------
        in_plane : array-like
            Desired in-plane coordinates for the point in the plane.
        offset : float
            Desired offset for the point.

        Returns
        -------
        output : array-like
            One point in the 3D-space, with given coordinates and offset.

        """
        return self.u * in_plane[0] + self.v * in_plane[1] + (offset -
                                                self.origin_offset) * self.w

    def intersection(self, l=None, other=None):
        """Function to compute the intersection with another line or plane.

        Parameters
        ----------
        l : Line
            Line intersecting the instance. (Default value = None)
        other : Plane
            Other plane. (Default value = None)

        Returns
        -------
        output : array-like or Line
            Point or line of intersection depending on input.

        """
        if l is not None:
            dir = l.get_direction()
            dot = np.dot(self.w, dir)
            if dot < 1e-10:
                return None
            p = l.point_at(0.0)
            k = - (self.origin_offset + np.dot(self.w, p)) / dot
            return p + k * dir
        else:
            dir = np.cross(self.w, other.w)
            if norm(dir) < self.tolerance:
                return None
            p = self.intersection_3_planes(self, other, Plane(normal=dir,
                tolerance=self.tolerance))
            return Line(p1=p, p2=p + dir, tolerance=self.tolerance)

    @classmethod
    def intersection_3_planes(self, p1, p2, p3):
        """Function to the compute the intersection of three planes, None if
        some planes are parallel.

        Parameters
        ----------
        p1 : Plane
            First plane.
        p2 : Plane
            Second plane.
        p3 : Plane
            Third plane.

        Returns
        -------
        output : array-like
            Point of intersection of three planes.

        """

        # Coefficients of the three planes linear equations.
        a1 = p1.w[0]
        b1 = p1.w[1]
        c1 = p1.w[2]
        d1 = p1.origin_offset

        a2 = p2.w[0]
        b2 = p2.w[1]
        c2 = p2.w[2]
        d2 = p2.origin_offset

        a3 = p3.w[0]
        b3 = p3.w[1]
        c3 = p3.w[2]
        d3 = p3.origin_offset

        # Direct Cramer resolution of the linear system.
        # This is still feasible for a 3x3 system.
        a23 = b2 * c3 - b3 * c2
        b23 = c2 * a3 - c3 * a2
        c23 = a2 * b3 - a3 * b2

        determinant = a1 * a23 + b1 * b23 + c1 * c23
        if abs(determinant) < 1e-10:
            return None

        r = 1.0 / determinant

        return np.array([
        (-a23 * d1 - (c1 * b3 - c3 * b1) * d2 - (c2 * b1 - c1 * b2) * d3) * r,
        (-b23 * d1 - (c3 * a1 - c1 * a3) * d2 - (c1 * a2 - c2 * a1) * d3) * r,
        (-c23 * d1 - (b1 * a3 - b3 * a1) * d2 - (b2 * a1 - b1 * a2) * d3) * r
        ])

    def contains(self, p):
        """Function to check if the instance contains a point.

        Parameters
        ----------
        p : array-like
            Point to check.

        Returns
        -------
        output : bool
            True if p belongs to plane, else False.

        """
        return abs(self.get_offset(point=p)) < self.tolerance

    def get_offset(self, point=None, plane=None):
        """Function to get the offset (oriented distance) of a parallel plane
        or a point.

        This method should be called only for parallel planes
        otherwise the result is not meaningful. The offset is 0 if both
        planes are the same, it is positive if the plane is on the plus side
        of the instance and negative if it is on the minus side, according to
        its natural orientation.

        Parameters
        ----------
        point : array-like
            Point to check. (Default value = None)
        plane : Plane
            Plane to check. (Default value = None)

        Returns
        -------
        output : float
            Offset of the plane or the point depending on input.

        """
        if plane is None:
            return np.dot(point, self.w) + self.origin_offset
        else:
            return self.origin_offset + (-plane.origin_offset if
                    self.same_orientation_as(plane) else plane.origin_offset)

    def same_orientation_as(self, other):
        """Function to check if the instance has the same orientation as another
        plane.

        Parameters
        ----------
        other : Plane
            Other plane to check against the instance.

        Returns
        -------
        output : bool
            True if the instance and the other plane have the same
            orientation, else False.

        """
        return np.dot(self.w, other.w) > 0.0


import inspect

from cheml.wrappers.database.unhooked import sklearn_db

file = open('sklearndb.py','w')

fns = [klass[0] for klass in inspect.getmembers(sklearn_db)]
extras = ['np','__builtins__', '__doc__', '__file__', '__name__', '__package__','mask','Input', 'Output', 'Parameter', 'req', 'regression_types', 'cv_types']
fns = [c for c in fns if c not in extras]
def add_parameter(block,item,default,format,description,options):
    line = "        %s = Parameter('%s', %s,'%s'," % (item, item, default, format)
    block.append(line)
    line = '                        description = "%s",'%description
    block.append(line)
    line = "                        options = %s)" %options
    block.append(line)
    return block


file.write('import numpy as np\n')
file.write('from .containers import Input, Output, Parameter, req, regression_types, cv_types\n\n')
for f in fns:
    block = []
    F = getattr(sklearn_db, f)
    line = "class %s(object):"%F.function
    block.append(line)
    line = "    task = '%s'"%F.task
    block.append(line)
    line = "    subtask = '%s'"%F.subtask
    block.append(line)
    line = "    host = '%s'"%F.host
    block.append(line)
    line = "    function = '%s'"%F.function
    block.append(line)
    line = "    modules = ('%s','%s')"%(F.modules[0],F.modules[1])
    block.append(line)
    line = "    requirements = (req(0), req(2))"
    block.append(line)
    line = "    documentation = " + '"' + F.documentation + '"'
    block.append(line)
    line = ""
    block.append(line)
    line = "    class Inputs:"
    block.append(line)
    if len(vars(F.Inputs).keys())>2:
        for item in vars(F.Inputs).keys():
            if item not in ('__module__','__doc__'):
                line = '        %s = Input("%s","%s", ('%(item, item,vars(F.Inputs)[item].short_description)
                for typ in vars(F.Inputs)[item].types:
                    line += '"%s",'%typ
                line += '))'
                block.append(line)
    else:
        line = "        pass"
        block.append(line)
    line = "    class Outputs:"
    block.append(line)
    if len(vars(F.Outputs).keys())>2:
        for item in vars(F.Outputs).keys():
            if item not in ('__module__', '__doc__'):
                line = '        %s = Output("%s","%s", ('%(item, item,vars(F.Outputs)[item].short_description)
                for typ in vars(F.Outputs)[item].types:
                    line += '"%s",'%typ
                line += '))'
                block.append(line)
    else:
        line = "        pass"
        block.append(line)
    line = "    class WParameters:"
    block.append(line)
    if len(vars(F.WParameters).keys())>2:
        for item in vars(F.WParameters).keys():
            if item not in ('__module__', '__doc__'):
                dic = vars(F.WParameters)[item]
                line = "        %s = Parameter('%s','%s','%s'," % (item, item, dic.default, dic.format)
                block.append(line)
                line = "                        description = " + '"%s",'%dic.description
                block.append(line)
                line = "                        options = %s)"%str(dic.options)
                block.append(line)
    else:
        line = "        pass"
        block.append(line)
    # Adding track_header to all functions
    description = "if True, the input dataframe's header will be transformed to the output dataframe"
    options = '(True, False)'
    block = add_parameter(block, 'track_header', 'True', 'Boolean', description, options)
    line = "    class FParameters:"
    block.append(line)
    line = "        pass"
    block.append(line)
    line = ""
    block.append(line)
    for line in block:
        file.write(line+'\n')
file.close()





import numpy as np
import sys
import os

class LookUpData:
    """Class to look up properties of elements stored in files.

    Attributes
    ----------
    element_ids : dict
        Dictionary containing element names (str) as the keys and index (int)
        in the periodic table as the values. Index is computed as Z - 1,
        where Z is the atomic number.
    element_names : array-like
        A list of abbreviated element names in the periodic table.
    sorting_order : array-like
        A list of electronegativity-based indices (int) used for printing the
        composition in a pretty format.
    element_order : dict
        Dictionary containing element names (str) as the keys and the sorting
        order (int) as values.
    all_properties : array-like
        A list of properties for which property-tables exist in
        magpie_python/lookup-data/.
    this_file_path : str
        Path to this file.
    abs_path : str
        Absolute path to the lookup-data directory.
    pair_abs_path : str
        Absolute path to the pair properties lookup directory.
    """

    # Element indices of the periodic table.
    element_ids = {"H": 0, "He": 1, "Li": 2, "Be": 3, "B": 4, "C": 5,
                   "N": 6, "O": 7, "F": 8, "Ne": 9, "Na": 10, "Mg": 11,
                   "Al": 12, "Si": 13, "P": 14, "S": 15, "Cl": 16,
                   "Ar": 17, "K": 18, "Ca": 19, "Sc": 20, "Ti": 21,
                   "V": 22, "Cr": 23, "Mn": 24, "Fe": 25, "Co": 26,
                   "Ni": 27, "Cu": 28, "Zn": 29, "Ga": 30, "Ge": 31,
                   "As": 32, "Se": 33, "Br": 34, "Kr": 35, "Rb": 36,
                   "Sr": 37, "Y": 38, "Zr": 39, "Nb": 40, "Mo": 41,
                   "Tc": 42, "Ru": 43, "Rh": 44, "Pd": 45, "Ag": 46,
                   "Cd": 47, "In": 48, "Sn": 49, "Sb": 50, "Te": 51,
                   "I": 52, "Xe": 53, "Cs": 54, "Ba": 55, "La": 56,
                   "Ce": 57, "Pr": 58, "Nd": 59, "Pm": 60, "Sm": 61,
                   "Eu": 62, "Gd": 63, "Tb": 64, "Dy": 65, "Ho": 66,
                   "Er": 67, "Tm": 68, "Yb": 69, "Lu": 70, "Hf": 71,
                   "Ta": 72, "W": 73, "Re": 74, "Os": 75, "Ir": 76,
                   "Pt": 77, "Au": 78, "Hg": 79, "Tl": 80, "Pb": 81,
                   "Bi": 82, "Po": 83, "At": 84, "Rn": 85, "Fr": 86,
                   "Ra": 87, "Ac": 88, "Th": 89, "Pa": 90, "U": 91,
                   "Np": 92, "Pu": 93, "Am": 94, "Cm": 95, "Bk": 96,
                   "Cf": 97, "Es": 98, "Fm": 99, "Md": 100, "No": 101,
                   "Lr": 102, "Rf": 103, "Db": 104, "Sg": 105,
                   "Bh": 106, "Hs": 107, "Mt": 108, "Ds": 109,
                   "Rg": 110, "Cn": 111}

    element_names = ["H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne",
                     "Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar", "K", "Ca",
                     "Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu",
                     "Zn", "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr",
                     "Y", "Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag",
                     "Cd", "In", "Sn", "Sb", "Te", "I", "Xe", "Cs", "Ba",
                     "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb",
                     "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Hf", "Ta", "W",
                     "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi",
                     "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th", "Pa", "U",
                     "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md",
                     "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds",
                     "Rg", "Cn"]

    sorting_order = [91, 1, 26, 62, 85, 102, 109, 111, 112, 2, 24, 53, 64,
                     74, 90, 104, 110, 3, 20, 27, 55, 60, 66, 68, 61, 72, 73,
                     78, 75, 67, 71, 83, 89, 103, 107, 108, 21, 25, 36, 54,
                     63, 88, 76, 92, 97, 93, 79, 69, 70, 80, 86, 87, 106,
                     105, 19, 22, 28, 30, 31, 32, 4, 33, 5, 34, 35, 37, 38,
                     39, 40, 6, 41, 43, 58, 100, 77, 94, 95, 98, 101, 81, 65,
                     99, 84, 82, 96, 7, 18, 23, 29, 44, 59, 57, 56, 42, 45,
                     46, 47, 48, 49, 50, 51, 52, 8, 9, 10, 11, 12, 13, 14,
                     15, 16, 17]

    element_order = {"H": 91, "He": 1, "Li": 26, "Be": 62, "B": 85, "C": 102,
                     "N": 109, "O": 111, "F": 112, "Ne": 2, "Na": 24,
                     "Mg": 53, "Al": 64, "Si": 74, "P": 90, "S": 104,
                     "Cl": 110, "Ar": 3, "K": 20, "Ca": 27, "Sc": 55,
                     "Ti": 60, "V": 66, "Cr": 68, "Mn": 61, "Fe": 72,
                     "Co": 73, "Ni": 78, "Cu": 75, "Zn": 67, "Ga": 71,
                     "Ge": 83, "As": 89, "Se": 103, "Br": 107, "Kr": 108,
                     "Rb": 21, "Sr": 25, "Y": 36, "Zr": 54, "Nb": 63,
                     "Mo": 88, "Tc": 76, "Ru": 92, "Rh": 97, "Pd": 93,
                     "Ag": 79, "Cd": 69, "In": 70, "Sn": 80, "Sb": 86,
                     "Te": 87, "I": 106, "Xe": 105, "Cs": 19, "Ba": 22,
                     "La": 28, "Ce": 30, "Pr": 31, "Nd": 32, "Pm": 4,
                     "Sm": 33, "Eu": 5, "Gd": 34, "Tb": 35, "Dy": 37,
                     "Ho": 38, "Er": 39, "Tm": 40, "Yb": 6, "Lu": 41,
                     "Hf": 43, "Ta": 58, "W": 100, "Re": 77, "Os": 94,
                     "Ir": 95, "Pt": 98, "Au": 101, "Hg": 81, "Tl": 65,
                     "Pb": 99, "Bi": 84, "Po": 82, "At": 96, "Rn": 7,
                     "Fr": 18, "Ra": 23, "Ac": 29, "Th": 44, "Pa": 59,
                     "U": 57, "Np": 56, "Pu": 42, "Am": 45, "Cm": 46,
                     "Bk": 47, "Cf": 48, "Es": 49, "Fm": 50, "Md": 51,
                     "No": 52, "Lr": 8, "Rf": 9, "Db": 10, "Sg": 11,
                     "Bh": 12, "Hs": 13, "Mt": 14, "Ds": 15, "Rg": 16,
                     "Cn": 17}

    # List of all the properties (except Abbreviation, IonizationEnergies,
    # OxidationStates) present in the directory: lookup-data/
    all_properties = ["AtomicVolume", "AtomicWeight",
                      "BoilingTemp", "BoilingT", "BulkModulus", "Column",
                      "CovalentRadius", "Density", "DipolePolarizability",
                      "ElectronAffinity", "Electronegativity",
                      "FirstIonizationEnergy", "FusionEnthalpy", "GSbandgap",
                      "GSenergy_pa", "GSestBCClatcnt", "GSestFCClatcnt",
                      "GSmagmom", "GSvolume_pa", "HeatCapacityMass",
                      "HeatCapacityMolar", "HeatFusion", "HHIp", "HHIr",
                      "ICSDVolume", "IsAlkali", "IsDBlock", "IsFBlock",
                      "IsMetalloid", "IsMetal", "IsNonmetal", "MeltingT",
                      "MendeleevNumber", "MiracleRadius", "NdUnfilled",
                      "NdValence", "NfUnfilled", "NfValence", "NpUnfilled",
                      "NpValence", "NsUnfilled", "NsValence", "Number",
                      "NUnfilled", "NValance", "n_ws^third", "phi",
                      "Polarizability", "Row", "ShearModulus",
                      "SpaceGroupNumber", "ZungerPP-r_d", "ZungerPP-r_pi",
                      "ZungerPP-r_p", "ZungerPP-r_sigma", "ZungerPP-r_s"]

    this_file_path = os.path.dirname(__file__)
    abs_path = os.path.join(this_file_path, "../../../lookup-data/")
    pair_abs_path = abs_path+"pair/"

    @classmethod
    def load_property(self, property):
        """Function to load a specific property from the directory containing
        all the lookup tables.

        Parameters
        ----------
        property : str
            Property whose values need to be loaded.

        Returns
        -------
        values : array-like
            A numpy array containing the property values for all the elements.

        Raises
        ------
        IOError
            If property table doesn't exist.

        """

        # IonizationEnergies and OxidationStates are 2-D arrays. So treat
        # them differently.
        if property == "IonizationEnergies" or property == "OxidationStates":
            return self.load_special_property(property)

        # Initialize the numpy array.
        values = np.zeros(len(self.element_ids), dtype=np.float)
        values.fill(np.nan)

        # Property file name.
        file = self.abs_path + property + ".table"
        try:
            prop_file = open(file, 'r')
        except IOError:
            raise IOError("File {} doesn't exist!!! Please make sure you " \
                  "specify the correct file name".format(file))

        else:
            for i in xrange(values.size):
                line = prop_file.readline().strip()
                try:
                    values[i] = float(line)
                except ValueError:
                    # Line is a string, which implies that data is missing.
                    # So we let the value of NaN stay.
                    continue
            prop_file.close()
        return values

    @classmethod
    def load_pair_property(self, property):
        """Function to load property of a binary system.

        Parameters
        ----------
        property : str
            Property whose values need to be loaded.

        Returns
        -------
        values : array-like
            A 2-D numpy array containing the property values for all the
            elements.

        Raises
        ------
        IOError
            If property table doesn't exist.

        """

        # Initialize the 2-D numpy array.
        values = np.zeros(len(self.element_ids), dtype=object)
        for i in xrange(len(self.element_ids)):
            values[i] = np.zeros(i, dtype=float)

        # Property file name.
        file = self.pair_abs_path + property + ".table"

        try:
            prop_file = open(file, 'r')
        except IOError:
            raise IOError("File {} doesn't exist!!! Please make sure you " \
                  "specify the correct file name".format(file))
        else:
            for line in prop_file.readlines():
                words = line.strip().split()
                if len(words) < 3:
                    continue
                else:
                    if words[0] not in self.element_ids or words[1] not in \
                            self.element_ids:
                        continue
                    elemA = self.element_ids[words[0]]
                    elemB = self.element_ids[words[1]]
                    if words[2].endswith("\n"):
                        print line
                        sys.exit(1)
                    values[max(elemA,elemB)][min(elemA,elemB)] = float(words[2])
            prop_file.close()
        return values

    @classmethod
    def load_pair_properties(self, properties):
        """Function to load multiple pair property values from the directory
        containing all the lookup tables.

        Parameters
        ----------
        properties : array-like
            A list of pair properties whose values need to be loaded.

        Returns
        -------
        values: dict
            A dictionary containing property name (str) as the key and a
            numpy array (float) containing the pair property values.

        """

        # Initialize the dictionary.
        values = {}

        for prop in properties:
            values[prop] = self.load_pair_property(prop)
        return values

    @classmethod
    def load_properties(self, properties):
        """Function to load multiple property values from the directory
        containing all the lookup tables.

        Parameters
        ----------
        properties : array-like
            A list of properties whose values need to be loaded.

        Returns
        -------
        values: dict
            A dictionary containing property name (str) as the key and a
            numpy array (float) containing the pair property values.

        """

        # Initialize the dictionary.
        values = {}

        for prop in properties:
            values[prop] = self.load_property(prop)
        return values

    @classmethod
    def load_special_property(self, property):
        """Function to load the special property files related to
        IonizationEnergies and OxidationStates.

        Parameters
        ----------
        property : str
            Property whose values need to be loaded.

        Returns
        -------
        values : array-like
            A 2-D numpy array containing the property values for all the
            elements.

        Raises
        ------
        IOError
            If property table doesn't exist.

        """
        # Property file name.
        file = self.abs_path + property + ".table"

        # Initialize the list.
        tmp_values = []

        try:
            prop_file = open(file, 'r')
        except IOError:
            print "File {} doesn't exist!!! Please make sure you " \
                  "specify the correct file name".format(file)
            sys.exit(1)
        else:
            for line in prop_file.readlines():
                words = line.strip().split()
                tmp_list = []
                for word in words:
                    tmp_list.append(float(word))
                tmp_values.append(tmp_list)
            prop_file.close()

        values = np.zeros(len(tmp_values), dtype=object)
        for i in xrange(len(tmp_values)):
            values[i] = np.asarray(tmp_values[i], dtype=float)
        return values


import unittest
import os
import numpy.testing as np_tst
from ....vassal.data.Atom import Atom
from ....vassal.data.Cell import Cell
from ....vassal.io.VASP5IO import VASP5IO

class testVASP5IO(unittest.TestCase):
    def test_conversion(self):
        # Make an FCC cell.
        cell = Cell()
        cell.set_basis(lengths=[3.5, 3.6, 3.4], angles=[89, 90, 91])
        cell.add_atom(Atom([0, 0, 0], 0))
        cell.add_atom(Atom([0.5, 0.5, 0], 1))
        cell.add_atom(Atom([0.5, 0, 0.5], 1))
        cell.add_atom(Atom([0, 0.5, 0.5], 1))
        cell.set_type_name(0, "Al")
        cell.set_type_name(1, "Ni")

        # Convert it to string.
        vio = VASP5IO()
        temp = vio.convert_structure_to_string(cell)

        # Convert it back.
        new_cell = vio.parse_file(list_of_lines=temp)

        # Check to make sure everything is good.
        self.assertAlmostEquals(cell.volume(), new_cell.volume(), delta=1e-4)
        self.assertEquals(cell.n_types(), new_cell.n_types())
        np_tst.assert_array_almost_equal(cell.get_lattice_vectors()[1],
                                         new_cell.get_lattice_vectors()[1],
                                         decimal=4)
        new_temp = vio.convert_structure_to_string(new_cell)
        np_tst.assert_equal(temp, new_temp)

    def test_parse_from_file(self):
        vio = VASP5IO()
        this_file_path = os.path.dirname(__file__)
        abs_path = os.path.join(this_file_path, "../../test-files/")
        cell = vio.parse_file(file_name=abs_path+"393-Ta1.vasp")
        self.assertAlmostEquals(556.549, cell.volume(), delta=1e-2)
        self.assertAlmostEquals(10.218, cell.get_lattice_vectors()[0][0],
                                delta=1e-2)
        self.assertEquals(30, cell.n_atoms())
        np_tst.assert_array_almost_equal([0.681, 0.818, 0.998],
                                         cell.get_atom(29).get_position(),
                                         decimal=2)


import types
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.utilities.filters.CompositionDistanceFilter import \
    CompositionDistanceFilter
from ....utility.tools.IonicCompoundFinder import IonicCompoundFinder

class IonicCompoundProximityAttributeGenerator:
    """Class to generate attributes based on the distance of a composition from
    a compositions that can form charge-neutral ionic compounds.

    Attributes
    ----------
    max_formula_unit : int
        Maximum number of atoms per formula unit.

    Notes
    -----
    This generator only computes a single feature: the L_1 distance between the
    composition of an entry and the nearest ionic compound (determined using
    IonicCompoundFinder). For a compound where it is not possible to form an
    ionic compound (e.g., only metallic elements), the entry is assigned
    arbitrarily large distance (equal to the number of elements in the alloy).
    The one adjustable parameter in this calculation is the maximum number of
    atoms per formula unit used when looking for ionic compounds. For binary
    compounds, the maximum conceivable number of elements in a formula unit
    is for a compound with a 9+ and a 5- species, which has 14 atoms in the
    formula unit. Consequently, we recommend using 14 or larger for this
    parameter.

    """

    # Maximum number of atoms per formula unit.
    max_formula_unit = 14

    def set_max_formula_unit(self, size):
        """Function to define the maximum number of atoms per formula unit.

        Parameters
        ----------
        size : int
            Desired size.

        """

        self.max_formula_unit = size

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CompositionEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        # Insert header names here.
        feat_headers.append("IonicCompoundDistance_MaxSize"+str(
            self.max_formula_unit))

        # Get ionic compound finder.
        finder = IonicCompoundFinder()
        finder.set_max_formula_unit_size(self.max_formula_unit)

        cdf = CompositionDistanceFilter()
        for entry in entries:
            elems = entry.get_element_ids()
            # Set the maximum distance to be equal to the number of elements.
            #  The maximum possible L_1 distance for an N-element system is N.
            finder.set_maximum_distance(len(elems))

            # If the material has only 1 element, set feature to 1.0.
            if len(elems) == 1:
                feat_values.append(1.0)
            else:
                # Get the list of all ionic compounds in the system.
                finder.set_nominal_composition(entry)
                ionic_compounds = finder.find_all_compounds()

                # Find the distance to the closest one. If no other
                # compounds, set distance to be the maximum possible.
                if not ionic_compounds:
                    feat_values.append(len(elems))
                else:
                    # print ionic_compounds[0]
                    # print cdf.compute_distance(entry, ionic_compounds[0], 1)
                    feat_values.append(cdf.compute_distance(entry,
                                        ionic_compounds[0], 1))

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


from itertools import combinations as comb
import numpy as np
from ....utility.EqualSumCombinations import EqualSumCombinations
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData

class PhaseDiagramCompositionEntryGenerator:
    """Class to generate composition entries at many points in many phase
    diagrams. Has two different ways of selecting compositions within phase
    diagrams.

    1. Even Spacing: Compositions are selected to be evenly-spaced throughout
    the phase diagram (e.g. A0.2B0.2C0.6, A0.4B0.2C0.4 etc.). This method is
    most appropriate for alloy systems.

    2. Simple Fractions: Compositions with the smallest denominator are
    selected (e.g. ABC, A2C, B2C, etc.). This method is most appropriate for
    phase diagrams that represent ordered crystalline compounds.

    Attributes
    ----------
    e_ids : array-like
        A list of int values denoting the list of elements to use.
    min_order : int
        Minimum number of constituents.
    max_order : int
        Maximum number of constituents.
    even_spacing : bool
        Whether to use even spacing or small integers.
    size : int
        Either number of stops in each direction or max denominator.

    """

    # List of elements to use (id is Z-1).
    e_ids = None

    # Minimum number of constituents.
    min_order = 1

    # Maximum number of constituents.
    max_order = 1

    # Whether to use even spacing or small integers.
    even_spacing = True

    # Either number of stops in each direction or max denominator.
    size = 3

    def set_elements_by_index(self, indices):
        """Function to define the list of elements to be included in the phase
        diagrams.

        Parameters
        ----------
        indices : array-like
            List of elements by index (Z-1).

        Raises
        ------
        ValueError
            If an element index is out of the allowed range.

        """
        self.e_ids = []
        for i in indices:
            if i < len(LookUpData.element_names):
                self.e_ids.append(i)
            else:
                raise ValueError("Index out of range: "+str(i))

    def set_elements_by_name(self, names):
        """Function to define the list of elements to be included in the phase
        diagrams.

        Parameters
        ----------
        names : array-like
            List of element names.

        Raises
        ------
        ValueError
            If an element name is not valid.

        """

        self.e_ids = []
        for name in names:
            if name in LookUpData.element_names:
                self.e_ids.append(LookUpData.element_names.index(name))
            else:
                raise ValueError("Element: "+name+" invalid!")

    def set_order(self, min_, max_):
        """Function to define the order of generated phase diagrams.

        Parameters
        ----------
        min_ : int
            Minimum number of constituents.
        max_ : int
            Maximum number of constituents.

        Raises
        ------
        ValueError
            If min_ or max_ is less than 1.

        """
        if min_ < 1 or max_ < 1:
            raise ValueError("Orders must be greater than 1.")
        self.min_order = min_
        self.max_order = max_

    def set_even_spacing(self, es):
        """Function to define whether to use evenly-spaced compositions (or
        low-denominator).

        Parameters
        ----------
        es : bool
            Boolean indicating the same.

        """
        self.even_spacing = es

    def set_size(self, size):
        """Function to define the number of points per binary diagram (in using
        even spacing) or the maximum denominator (for low-denominator).

        Parameters
        ----------
        size : int
            Desired size parameter.

        Raises
        ------
        ValueError
            If input is less than 2.

        """
        if size < 2:
            raise ValueError("Size must be greater than 1.")
        self.size = size

    def generate_alloy_compositions(self):
        """Function to generate evenly-spaced compositions. Generates
        compositions for all diagrams up to the user-specified Minimum order.

        For example: If the user wants ternary diagrams with a minimum
        spacing of 0.25 this code will generate the following map:
        1 -> ([1.0])
        2 -> ([0.25, 0.75], [0.5, 0.5], [0.75, 0.25])
        3 -> ([0.5, 0.25, 0.25], [0.25, 0.5, 0.25], [0.25, 0.25, 0.5])

        Returns
        -------
        output : dict
            A dictionary containing <order, possible compositions> as <key,
            value> pairs. Here, order is the number of elements and possible
            compositions is a list of numpy arrays containing the fractions
            of elements.

        """

        output = {}

        # Add in diagrams of greater order.
        for order in range(self.min_order, self.max_order+1):
            if order == 1:
                tmp_list = []
                tmp_list.append(np.array([1.0]))
                output[order] = tmp_list
                continue

            tmp_list = []
            es = EqualSumCombinations(self.size-1, order)
            for compI in es.get_combinations(self.size - 1, order):
                if 0 in compI:
                    # Don't add compositions from a lower-order diagram.
                    continue
                comp = np.zeros(order)
                for i in range(order):
                    comp[i] = compI[i]/ float(self.size - 1.0)
                tmp_list.append(comp)
            output[order] = tmp_list
        return output

    def generate_crystal_compositions(self):
        """Function to generate compositions with small denominators. Generates
        compositions for all diagrams up to the user-specified Minimum order.

        For example: If the user wants ternary diagrams with a maximum
        denominator of 3 this code will generate the following map:
        1 -> ([1])
        2 -> ([1/3, 2/3], [1/2, 1/2], [2/3, 1/3])
        3 -> ([1/3, 1/3, 1/3])

        Returns
        -------
        output : dict
            A dictionary containing <order, possible compositions> as <key,
            value> pairs. Here, order is the number of elements and possible
            compositions is a list of numpy arrays containing the fractions
            of elements.

        """

        output = {}

        # Add in diagrams of greater order.
        for order in range(self.min_order, self.max_order+1):
            if order == 1:
                tmp_list = []
                tmp_list.append(np.array([1.0]))
                output[order] = tmp_list
                continue

            tmp_list = []
            reduced_examples = []
            for d in range(order, self.size+1):
                es = EqualSumCombinations(d, order)
                for compI in es.get_combinations(d, order):
                    if 0 in compI:
                        # Don't add compositions from a lower-order diagram.
                        continue
                    comp = np.zeros(order)
                    red_comp = np.zeros(order)
                    for i in range(order):
                        comp[i] = float(compI[i])
                        red_comp[i] = comp[i] / d

                    # Check if this composition is already represented.
                    was_found = False
                    for ex in reduced_examples:
                        if np.array_equal(red_comp, ex):
                            was_found = True
                            break

                    if not was_found:
                        tmp_list.append(comp)
                        reduced_examples.append(red_comp)
            output[order] = tmp_list
        return output

    def generate_entries(self):
        """Function to generate the list of entries corresponding to the list of
        compositions, element names specified by the user and the mapping of
        number of elements to compositions.

        Returns
        -------
        entries: array-like
            A list of CompositionEntry's.

        """
        entries = []

        # Get the correct composition mapping.
        compositions = self.generate_alloy_compositions() if \
            self.even_spacing else self.generate_crystal_compositions()

        for order,list_of_fractions in compositions.iteritems():

            # Generate all possible combinations of elements of a given order.
            compounds = [list(i) for i in comb(self.e_ids, order)]
            # print order, len(compounds)*len(list_of_fractions)
            for frac in list_of_fractions:
                for compound in compounds:
                    entry = CompositionEntry(element_ids=compound,
                                             fractions=list(frac))
                    entries.append(entry)
        return entries


"""
The 'cheml.utils' module includes list_del_indices, std_datetime_str, slurm_script_exclusive,
isfloat, string2nan, 
last modified date: April 25, 2016
"""

from .utilities import list_del_indices
from .utilities import std_datetime_str
from .utilities import slurm_script_exclusive
from .utilities import chunk
from .utilities import choice
from .utilities import return2Dshape
from .utilities import bool_formatter

from .validation import string2nan
from .validation import isfloat
from .validation import islist
from .validation import istuple
from .validation import isnpdot
from .validation import isint
from .validation import value
from .validation import check_input
from .validation import check_object_col

from .lib_jcode import tot_exec_time_str

__all__ = [
    'list_del_indices',
    'std_datetime_str',
    'slurm_script_exclusive',
    'isfloat',
    'string2nan',
]



from ....attributes.generators.crystal\
    .LocalPropertyDifferenceAttributeGenerator import \
    LocalPropertyDifferenceAttributeGenerator

class LocalPropertyVarianceAttributeGenerator(
    LocalPropertyDifferenceAttributeGenerator):
    """Class to compute attributes based on the local variance in elemental
    properties around each atom.

    See Also
    --------
    LocalPropertyDifferenceAttributeGenerator : Super class of this class.

    """

    def __init__(self, shells=None):
        """Function to create instance and initialize fields.

        Parameters
        ----------
        shells : array-like
            Shells to be considered. A list of int values.

        """

        LocalPropertyDifferenceAttributeGenerator.__init__(self, shells)
        self.attr_name = "NeighVar"

    def get_atom_properties(self, voro, shell, prop_values):
        """Function to compute the properties of a certain neighbor cell for
        each atom, given the Voronoi tessellation and properties of each atom
        type.

        Parameters
        ----------
        voro : VoronoiCellBasedAnalysis
            Analysis tool.
        shell : int
            Index of shell.
        prop_values : array-like
            Properties of each atom type. A list or NumPy array of float values.

        Returns
        -------
        output : array-like
            Properties of each atom. A list or NumPy array of float values.

        """

        output = voro.neighbor_property_variances(prop_values, shell)
        return output


import unittest
import numpy.testing as np_tst
from .....vassal.analysis.voronoi.VoronoiEdge import VoronoiEdge
from .....vassal.analysis.voronoi.VoronoiFace import VoronoiFace
from .....vassal.data.Atom import Atom
from .....vassal.data.AtomImage import AtomImage
from .....vassal.data.Cell import Cell

class testVoronoiEdge(unittest.TestCase):
    def setUp(self):
        # Make a simple crystal.
        self.cell = Cell()
        self.cell.add_atom(Atom([0, 0, 0], 0))

    def tearDown(self):
        self.cell = None

    def test_initialize(self):
        # Initialize faces.
        image = AtomImage(self.cell.get_atom(0), [0, 0, 1])
        face1 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [0, 1, 0])
        face2 = VoronoiFace(self.cell.get_atom(0), image, radical=True)

        # Create edge.
        edge = VoronoiEdge(face1, face2)

        # Check out properties.
        np_tst.assert_array_almost_equal([-1, 0, 0], edge.get_line(
        ).get_direction())

    def test_is_next(self):
        # Initialize faces.
        image = AtomImage(self.cell.get_atom(0), [0, 0, 1])
        face1 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [0, 1, 0])
        face2 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [1, 0, 0])
        face3 = VoronoiFace(self.cell.get_atom(0), image, radical=True)

        # Create edges.
        edge1 = VoronoiEdge(face1, face2)
        edge2 = VoronoiEdge(face1, face3)

        # Check proper ordering.
        self.assertTrue(edge2.is_ccw(edge2=edge1))
        self.assertFalse(edge1.is_ccw(edge2=edge2))
        self.assertFalse(edge2.is_ccw(edge2=edge2))

    def test_simple_intersection(self):
        # Initialize faces.
        image = AtomImage(self.cell.get_atom(0), [0, 0, 1])
        face1 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [0, 1, 0])
        face2 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [1, 0, 0])
        face3 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [2, 0, 0])
        face4 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [-1, 0, 0])
        face5 = VoronoiFace(self.cell.get_atom(0), image, radical=True)

        # Create edges.
        edge1 = VoronoiEdge(face1, face2)
        edge2 = VoronoiEdge(face1, face3)
        edge3 = VoronoiEdge(face1, face4)
        edge4 = VoronoiEdge(face1, face5)

        # Compute intersection.
        self.assertTrue(VoronoiEdge.compute_intersection(edge1, edge3))
        self.assertTrue(VoronoiEdge.compute_intersection(edge1, edge2))
        self.assertTrue(VoronoiEdge.compute_intersection(edge1, edge4))

        # Test properties.
        np_tst.assert_array_almost_equal([0.5, 0.5, 0.5],
                        edge1.get_start_vertex().get_position())
        np_tst.assert_array_almost_equal([0.5, 0.5, 0.5],
                        edge2.get_end_vertex().get_position())
        self.assertAlmostEquals(1.0, edge1.get_length(), delta=1e-6)

        # Compute intersection that shouldn't happen.
        self.assertFalse(VoronoiEdge.compute_intersection(edge1, edge3))

    def test_join(self):
        # Initialize faces.
        image = AtomImage(self.cell.get_atom(0), [0, 0, 1])
        face1 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [0, 1, 0])
        face2 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [1, 0, 0])
        face3 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [2, 0, 0])
        face4 = VoronoiFace(self.cell.get_atom(0), image, radical=True)

        # Create edges.
        edge1 = VoronoiEdge(face1, face2)
        edge2 = VoronoiEdge(face1, face3)
        edge3 = VoronoiEdge(face1, face4)

        # Join edge1 & edge2.
        VoronoiEdge.compute_intersection(edge1, edge2, just_join=True)
        np_tst.assert_array_almost_equal([0.5, 0.5, 0.5],
                        edge1.get_start_vertex().get_position())
        self.assertEquals(edge2, edge1.get_previous_edge())

        # Join edge1 & edge3.
        VoronoiEdge.compute_intersection(edge1, edge3, just_join=True)
        np_tst.assert_array_almost_equal([1.0, 0.5, 0.5],
                    edge1.get_start_vertex().get_position())
        self.assertEquals(edge3, edge1.get_previous_edge())

    def test_compare(self):
        # Initialize faces.
        image = AtomImage(self.cell.get_atom(0), [0, 0, 1])
        face1 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [0, 1, 0])
        face2 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [1, 0, 0])
        face3 = VoronoiFace(self.cell.get_atom(0), image, radical=True)

        # Create edges.
        edge1 = VoronoiEdge(face1, face2)
        edge2 = VoronoiEdge(face1, face3)

        # Tests.
        self.assertEquals(0, edge1.__cmp__(edge1))
        self.assertTrue(edge1.__cmp__(edge2) != 0)
        self.assertTrue(edge1.__cmp__(edge2) == -1 * edge2.__cmp__(edge1))
        self.assertFalse(edge1.__eq__(edge2))
        self.assertTrue(edge1.__eq__(edge1))

    def test_next_edge(self):
        # Initialize faces.
        image = AtomImage(self.cell.get_atom(0), [0, 0, 1])
        face1 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [0, 1, 0])
        face2 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [-1, 1, 0])
        face3 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [-1, 0, 0])
        face4 = VoronoiFace(self.cell.get_atom(0), image, radical=True)

        # Create edges.
        edge1 = VoronoiEdge(face1, face2)
        edge2 = VoronoiEdge(face1, face3)
        edge3 = VoronoiEdge(face1, face4)

        # Verify that edge2 and edge3 intersect edge1 in the same place.
        p1 = edge1.get_line().intersection(edge2.get_line())
        p2 = edge1.get_line().intersection(edge3.get_line())
        np_tst.assert_array_almost_equal(p1, p2)

        # Verify that when tasked with distinguishing between the two, it
        # recognizes that edge3 is the correct choice
        choices = [edge2, edge3]
        self.assertTrue(edge1.is_ccw(edge2=edge2))
        self.assertEquals(edge3, edge1.find_next_edge(choices))

        # Add more conflicting choices.
        image = AtomImage(self.cell.get_atom(0), [-2, 0, 0])
        face4 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        # A farther edge.
        choices.append(VoronoiEdge(face1, face4))
        image = AtomImage(self.cell.get_atom(0), [1, 0, 0])
        face4 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        # A CW edge.
        choices.append(VoronoiEdge(face1, face4))

        self.assertEquals(edge3, edge1.find_next_edge(choices))

    def test_pair(self):
        # Initialize faces.
        image = AtomImage(self.cell.get_atom(0), [0, 0, 1])
        face1 = VoronoiFace(self.cell.get_atom(0), image, radical=True)
        image = AtomImage(self.cell.get_atom(0), [0, 1, 0])
        face2 = VoronoiFace(self.cell.get_atom(0), image, radical=True)

        # Create edge.
        edge = VoronoiEdge(face1, face2)

        # Create pair.
        pair = edge.generate_pair()

        # Tests.
        self.assertEquals(edge.get_edge_face(), pair.get_intersecting_face())
        self.assertEquals(edge.get_intersecting_face(), pair.get_edge_face())
        dir1 = edge.get_line().get_direction()
        dir2 = -1 * pair.get_line().get_direction()
        np_tst.assert_array_almost_equal(dir1, dir2)


import numpy as np
from .containers import Input, Output, Parameter, req, regression_types, cv_types

class Binarizer(object):
    task = 'Prepare'
    subtask = 'feature representation'
    host = 'sklearn'
    function = 'Binarizer'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Binarizer.html#sklearn.preprocessing.Binarizer"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's Binarizer class", ("<class 'sklearn.preprocessing.data.Binarizer'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's Binarizer class", ("<class 'sklearn.preprocessing.data.Binarizer'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class PolynomialFeatures(object):
    task = 'Prepare'
    subtask = 'feature representation'
    host = 'sklearn'
    function = 'PolynomialFeatures'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.PolynomialFeatures.html#sklearn.preprocessing.PolynomialFeatures"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's PolynomialFeatures class", ("<class 'sklearn.preprocessing.data.PolynomialFeatures'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's PolynomialFeatures class", ("<class 'sklearn.preprocessing.data.PolynomialFeatures'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class OneHotEncoder(object):
    task = 'Prepare'
    subtask = 'feature representation'
    host = 'sklearn'
    function = 'OneHotEncoder'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html#sklearn.preprocessing.OneHotEncoder"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's OneHotEncoder class", ("<class 'sklearn.preprocessing.data.OneHotEncoder'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's OneHotEncoder class", ("<class 'sklearn.preprocessing.data.OneHotEncoder'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass

class Imputer(object):
    task = 'Prepare'
    subtask = 'preprocessor'
    host = 'sklearn'
    function = 'Imputer'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Imputer.html#sklearn.preprocessing.Imputer"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's Imputer class", ("<class 'sklearn.preprocessing.imputation.Imputer'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's Imputer class", ("<class 'sklearn.preprocessing.imputation.Imputer'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass

class KernelPCA(object):
    task = 'Prepare'
    subtask = 'feature transformation'
    host = 'sklearn'
    function = 'KernelPCA'
    modules = ('sklearn','decomposition')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.KernelPCA.html#sklearn.decomposition.KernelPCA"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's KernelPCA class", ("<class 'sklearn.decomposition.kernel_pca.KernelPCA'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's KernelPCA class", ("<class 'sklearn.decomposition.kernel_pca.KernelPCA'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', False,'Boolean',
                        description = "Always False, the header of input dataframe is not equivalent with the transformed dataframe",
                        options = (False))
    class FParameters:
        pass
class PCA(object):
    task = 'Prepare'
    subtask = 'feature transformation'
    host = 'sklearn'
    function = 'PCA'
    modules = ('sklearn','decomposition')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html#sklearn.decomposition.PCA"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's PCA class", ("<class 'sklearn.decomposition.pca.PCA'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's PCA class", ("<class 'sklearn.decomposition.pca.PCA'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', False,'Boolean',
                        description = "Always False, the header of input dataframe is not equivalent with the transformed dataframe",
                        options = (False))
    class FParameters:
        pass

class Normalizer(object):
    task = 'Prepare'
    subtask = 'scale'
    host = 'sklearn'
    function = 'Normalizer'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Normalizer.html#sklearn.preprocessing.Normalizer"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's Normalizer class", ("<class 'sklearn.preprocessing.data.Normalizer'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's Normalizer class", ("<class 'sklearn.preprocessing.data.Normalizer'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api None: only make a new api ",
                        options = ('fit_transform', 'transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class StandardScaler(object):
    task = 'Prepare'
    subtask = 'scale'
    host = 'sklearn'
    function = 'StandardScaler'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html#sklearn.preprocessing.StandardScaler"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's StandardScaler class", ("<class 'sklearn.preprocessing.data.StandardScaler'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's StandardScaler class", ("<class 'sklearn.preprocessing.data.StandardScaler'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class MinMaxScaler(object):
    task = 'Prepare'
    subtask = 'scale'
    host = 'sklearn'
    function = 'MinMaxScaler'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html#sklearn.preprocessing.MinMaxScaler"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's MinMaxScaler class", ("<class 'sklearn.preprocessing.data.MinMaxScaler'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's MinMaxScaler class", ("<class 'sklearn.preprocessing.data.MinMaxScaler'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class MaxAbsScaler(object):
    task = 'Prepare'
    subtask = 'scale'
    host = 'sklearn'
    function = 'MaxAbsScaler'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MaxAbsScaler.html#sklearn.preprocessing.MaxAbsScaler"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's MaxAbsScaler class", ("<class 'sklearn.preprocessing.data.MaxAbsScaler'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's MaxAbsScaler class", ("<class 'sklearn.preprocessing.data.MaxAbsScaler'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class RobustScaler(object):
    task = 'Prepare'
    subtask = 'scale'
    host = 'sklearn'
    function = 'RobustScaler'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html#sklearn.preprocessing.RobustScaler"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's RobustScaler class", ("<class 'sklearn.preprocessing.data.RobustScaler'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's RobustScaler class", ("<class 'sklearn.preprocessing.data.RobustScaler'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass

class ShuffleSplit(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'ShuffleSplit'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.ShuffleSplit.html#sklearn.model_selection.ShuffleSplit"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        fold_gen = Output("fold_gen","Generator of indices to split data into training and test set", ("<type 'generator'>",))
        api = Output("api","instance of scikit-learn's ShuffleSplit class", ("<class 'sklearn.model_selection._split.ShuffleSplit'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('split', None))
    class FParameters:
        pass
class StratifiedShuffleSplit(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'StratifiedShuffleSplit'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.StratifiedShuffleSplit.html#sklearn.model_selection.StratifiedShuffleSplit"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        fold_gen = Output("fold_gen","Generator of indices to split data into training and test set", ("<type 'generator'>",))
        api = Output("api","instance of scikit-learn's StratifiedShuffleSplit class", ("<class 'sklearn.model_selection._split.StratifiedShuffleSplit'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('split', None))
    class FParameters:
        pass
class train_test_split(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'train_test_split'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        dfx_test = Output("dfx_test","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfy_train = Output("dfy_train","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfy_test = Output("dfy_test","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx_train = Output("dfx_train","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class KFold(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'KFold'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.KFold.html"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's KFold class", ("<class 'sklearn.model_selection._split.KFold'>",))
        fold_gen = Output("fold_gen","Generator of indices to split data into training and test set", ("<type 'generator'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('split', None))
    class FParameters:
        pass

class ARDRegression(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'ARDRegression'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ARDRegression.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's ARDRegression class", ("<class 'sklearn.linear_model.bayes.ARDRegression'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's ARDRegression class", ("<class 'sklearn.linear_model.bayes.ARDRegression'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class BayesianRidge(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'BayesianRidge'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.BayesianRidge.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's BayesianRidge class", ("<class 'sklearn.linear_model.bayes.BayesianRidge'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's BayesianRidge class", ("<class 'sklearn.linear_model.bayes.BayesianRidge'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class ElasticNet(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'ElasticNet'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ElasticNet.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's ElasticNet class", ("<class 'sklearn.linear_model.coordinate_descent.ElasticNet'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's ElasticNet class", ("<class 'sklearn.linear_model.coordinate_descent.ElasticNet'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class KernelRidge(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'KernelRidge'
    modules = ('sklearn','kernel_ridge')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.kernel_ridge.KernelRidge.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's KernelRidge class", ("<class 'sklearn.kernel_ridge.KernelRidge'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's KernelRidge class", ("<class 'sklearn.kernel_ridge.KernelRidge'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class Lars(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'Lars'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lars.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's Lars class", ("<class 'sklearn.linear_model.least_angle.Lars'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's Lars class", ("<class 'sklearn.linear_model.least_angle.Lars'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class Lasso(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'Lasso'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lasso.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's Lasso class", ("<class 'sklearn.linear_model.coordinate_descent.Lasso'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's Lasso class", ("<class 'sklearn.linear_model.coordinate_descent.Lasso'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class LassoLars(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LassoLars'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoLars.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's LassoLars class", ("<class 'sklearn.linear_model.least_angle.LassoLars'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's LassoLars class", ("<class 'sklearn.linear_model.least_angle.LassoLars'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class LinearRegression(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LinearRegression'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's LinearRegression class", ("<class 'sklearn.linear_model.base.LinearRegression'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's LinearRegression class", ("<class 'sklearn.linear_model.base.LinearRegression'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class LinearSVR(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LinearSVR'
    modules = ('sklearn','svm')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.svm.LinearSVR.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's LinearSVR class", ("<class 'sklearn.svm.classes.LinearSVR'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's LinearSVR class", ("<class 'sklearn.svm.classes.LinearSVR'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class LogisticRegression(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LogisticRegression'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's LogisticRegression class", ("<class 'sklearn.linear_model.logistic.LogisticRegression'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's LogisticRegression class", ("<class 'sklearn.linear_model.logistic.LogisticRegression'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class MLPRegressor(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'MLPRegressor'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPRegressor.html#sklearn.neural_network.MLPRegressor"

    class Inputs:
        api = Input("api","instance of scikit-learn's MLPRegressor class", ("<class 'sklearn.neural_network.multilayer_perceptron.MLPRegressor'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's MLPRegressor class", ("<class 'sklearn.neural_network.multilayer_perceptron.MLPRegressor'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class MultiTaskElasticNet(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'MultiTaskElasticNet'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.MultiTaskElasticNet.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's MultiTaskElasticNet class", ("<class 'sklearn.linear_model.coordinate_descent.MultiTaskElasticNet'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's MultiTaskElasticNet class", ("<class 'sklearn.linear_model.coordinate_descent.MultiTaskElasticNet'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class MultiTaskLasso(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'MultiTaskLasso'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.MultiTaskLasso.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's MultiTaskLasso class", ("<class 'sklearn.linear_model.coordinate_descent.MultiTaskLasso'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's MultiTaskLasso class", ("<class 'sklearn.linear_model.coordinate_descent.MultiTaskLasso'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class NuSVR(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'NuSVR'
    modules = ('sklearn','svm')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.svm.NuSVR.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's NuSVR class", ("<class 'sklearn.svm.classes.NuSVR'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's NuSVR class", ("<class 'sklearn.svm.classes.NuSVR'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class Ridge(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'Ridge'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's Ridge class", ("<class 'sklearn.linear_model.ridge.Ridge'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's Ridge class", ("<class 'sklearn.linear_model.ridge.Ridge'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class SGDRegressor(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'SGDRegressor'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDRegressor.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's SGDRegressor class", ("<class 'sklearn.linear_model.stochastic_gradient.SGDRegressor'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's SGDRegressor class", ("<class 'sklearn.linear_model.stochastic_gradient.SGDRegressor'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class SVR(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'SVR'
    modules = ('sklearn','svm')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.svm.SVR.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's SVR class", ("<class 'sklearn.svm.classes.SVR'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's SVR class", ("<class 'sklearn.svm.classes.SVR'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass

class GridSearchCV(object):
    task = 'Search'
    subtask = 'grid'
    host = 'sklearn'
    function = 'GridSearchCV'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        estimator = Input("estimator","instance of a machine learning class", tuple(["<type 'str'>"]+list(regression_types())))
        scorer = Input("scorer","instance of scikit-learn's make_scorer class", ("<type 'str'>","<class 'sklearn.metrics.scorer._PredictScorer'>",))
    class Outputs:
        cv_results_ = Output("cv_results_","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's GridSearchCV class", ("<class 'sklearn.grid_search.GridSearchCV'>",))
        best_estimator_ = Output("best_estimator_","instance of a machine learning class", regression_types())
    class WParameters:
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class cross_val_predict(object):
    task = 'Search'
    subtask = 'validate'
    host = 'sklearn'
    function = 'cross_val_predict'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.cross_val_predict.html#sklearn.model_selection.cross_val_predict"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        estimator = Input("estimator","instance of a machine learning class", ("<class 'sklearn.linear_model.base.LinearRegression'>","<class 'sklearn.linear_model.ridge.Ridge'>","<class 'sklearn.kernel_ridge.KernelRidge'>","<class 'sklearn.linear_model.coordinate_descent.Lasso'>","<class 'sklearn.linear_model.coordinate_descent.MultiTaskLasso'>","<class 'sklearn.linear_model.coordinate_descent.ElasticNet'>","<class 'sklearn.linear_model.coordinate_descent.MultiTaskElasticNet'>","<class 'sklearn.linear_model.least_angle.Lars'>","<class 'sklearn.linear_model.least_angle.LassoLars'>","<class 'sklearn.linear_model.bayes.BayesianRidge'>","<class 'sklearn.linear_model.bayes.ARDRegression'>","<class 'sklearn.linear_model.logistic.LogisticRegression'>","<class 'sklearn.linear_model.stochastic_gradient.SGDRegressor'>","<class 'sklearn.svm.classes.SVR'>","<class 'sklearn.svm.classes.NuSVR'>","<class 'sklearn.svm.classes.LinearSVR'>","<class 'sklearn.neural_network.multilayer_perceptron.MLPRegressor'>",))
        scorer = Input("scorer","instance of scikit-learn's make_scorer class", ("<class 'sklearn.metrics.scorer._PredictScorer'>",))
    class Outputs:
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class cross_val_score(object):
    task = 'Search'
    subtask = 'validate'
    host = 'sklearn'
    function = 'cross_val_score'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.cross_val_score.html#sklearn.model_selection.cross_val_score"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        estimator = Input("estimator","instance of a machine learning class", ("<class 'sklearn.linear_model.base.LinearRegression'>","<class 'sklearn.linear_model.ridge.Ridge'>","<class 'sklearn.kernel_ridge.KernelRidge'>","<class 'sklearn.linear_model.coordinate_descent.Lasso'>","<class 'sklearn.linear_model.coordinate_descent.MultiTaskLasso'>","<class 'sklearn.linear_model.coordinate_descent.ElasticNet'>","<class 'sklearn.linear_model.coordinate_descent.MultiTaskElasticNet'>","<class 'sklearn.linear_model.least_angle.Lars'>","<class 'sklearn.linear_model.least_angle.LassoLars'>","<class 'sklearn.linear_model.bayes.BayesianRidge'>","<class 'sklearn.linear_model.bayes.ARDRegression'>","<class 'sklearn.linear_model.logistic.LogisticRegression'>","<class 'sklearn.linear_model.stochastic_gradient.SGDRegressor'>","<class 'sklearn.svm.classes.SVR'>","<class 'sklearn.svm.classes.NuSVR'>","<class 'sklearn.svm.classes.LinearSVR'>","<class 'sklearn.neural_network.multilayer_perceptron.MLPRegressor'>",))
        scorer = Input("scorer","instance of scikit-learn's make_scorer class", ("<class 'sklearn.metrics.scorer._PredictScorer'>",))
    class Outputs:
        scores = Output("scores","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class learning_curve(object):
    task = 'Search'
    subtask = 'grid'
    host = 'sklearn'
    function = 'learning_curve'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.learning_curve.html#sklearn.model_selection.learning_curve"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        estimator = Input("estimator","instance of a machine learning class", ("<class 'sklearn.linear_model.base.LinearRegression'>","<class 'sklearn.linear_model.ridge.Ridge'>","<class 'sklearn.kernel_ridge.KernelRidge'>","<class 'sklearn.linear_model.coordinate_descent.Lasso'>","<class 'sklearn.linear_model.coordinate_descent.MultiTaskLasso'>","<class 'sklearn.linear_model.coordinate_descent.ElasticNet'>","<class 'sklearn.linear_model.coordinate_descent.MultiTaskElasticNet'>","<class 'sklearn.linear_model.least_angle.Lars'>","<class 'sklearn.linear_model.least_angle.LassoLars'>","<class 'sklearn.linear_model.bayes.BayesianRidge'>","<class 'sklearn.linear_model.bayes.ARDRegression'>","<class 'sklearn.linear_model.logistic.LogisticRegression'>","<class 'sklearn.linear_model.stochastic_gradient.SGDRegressor'>","<class 'sklearn.svm.classes.SVR'>","<class 'sklearn.svm.classes.NuSVR'>","<class 'sklearn.svm.classes.LinearSVR'>","<class 'sklearn.neural_network.multilayer_perceptron.MLPRegressor'>",))
        scorer = Input("scorer","instance of scikit-learn's make_scorer class", ("<class 'sklearn.metrics.scorer._PredictScorer'>",))
        cv = Input("cv","instance of scikit-learn's cross validation generator", ("<class 'sklearn.model_selection._split.KFold'>","<class 'sklearn.model_selection._split.ShuffleSplit'>","<class 'sklearn.model_selection._split.StratifiedShuffleSplit'>",))
    class Outputs:
        train_sizes_abs = Output("train_sizes_abs","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        extended_result_ = Output("extended_result_","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        test_scores = Output("test_scores","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        train_scores = Output("train_scores","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class evaluate_regression(object):
    task = 'Search'
    subtask = 'evaluate'
    host = 'sklearn'
    function = 'evaluate_regression'
    modules = ('sklearn','metrics')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/dev/modules/model_evaluation.html#regression-metrics"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfy_predict = Input("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        evaluation_results_ = Output("evaluation_results_","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        evaluator = Output("evaluator","dictionary of metrics and their score function", ("<type 'dict'>",))
    class WParameters:
        mae_multioutput = Parameter('mae_multioutput','uniform_average','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html#sklearn.metrics.mean_absolute_error',
                        description = "",
                        options = ('raw_values', 'uniform_average'))
        r2_score = Parameter('r2_score','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html#sklearn.metrics.r2_score',
                        description = "",
                        options = (True, False))
        mean_absolute_error = Parameter('mean_absolute_error','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html#sklearn.metrics.mean_absolute_error',
                        description = "",
                        options = (True, False))
        multioutput = Parameter('multioutput','uniform_average','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html#sklearn.metrics.r2_score',
                        description = "",
                        options = ('raw_values', 'uniform_average', 'variance_weighted'))
        mse_sample_weight = Parameter('mse_sample_weight','None','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = [])
        rmse_multioutput = Parameter('rmse_multioutput','uniform_average','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = ('raw_values', 'uniform_average'))
        median_absolute_error = Parameter('median_absolute_error','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.median_absolute_error.html#sklearn.metrics.median_absolute_error',
                        description = "",
                        options = (True, False))
        mae_sample_weight = Parameter('mae_sample_weight','None','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html#sklearn.metrics.mean_absolute_error',
                        description = "",
                        options = [])
        rmse_sample_weight = Parameter('rmse_sample_weight','None','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = [])
        mean_squared_error = Parameter('mean_squared_error','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = (True, False))
        root_mean_squared_error = Parameter('root_mean_squared_error','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = (True, False))
        explained_variance_score = Parameter('explained_variance_score','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.explained_variance_score.html#sklearn.metrics.explained_variance_score',
                        description = "",
                        options = (True, False))
        r2_sample_weight = Parameter('r2_sample_weight','None','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html#sklearn.metrics.r2_score',
                        description = "",
                        options = [])
        ev_sample_weight = Parameter('ev_sample_weight','None','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.explained_variance_score.html#sklearn.metrics.explained_variance_score',
                        description = "",
                        options = [])
        ev_multioutput = Parameter('ev_multioutput','uniform_average','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.explained_variance_score.html#sklearn.metrics.explained_variance_score',
                        description = "",
                        options = ('raw_values', 'uniform_average', 'variance_weighted'))
        mse_multioutput = Parameter('mse_multioutput','uniform_average','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = ('raw_values', 'uniform_average'))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class scorer_regression(object):
    task = 'Search'
    subtask = 'evaluate'
    host = 'sklearn'
    function = 'scorer_regression'
    modules = ('sklearn','metrics')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/0.15/modules/generated/sklearn.metrics.make_scorer.html#sklearn.metrics.make_scorer"

    class Inputs:
        pass

    class Outputs:
        scorer = Output("scorer","Callable object that returns a scalar score", ("<class 'sklearn.metrics.scorer._PredictScorer'>",))

    class WParameters:
        metric = Parameter('metric','mae','http://scikit-learn.org/dev/modules/model_evaluation.html#regression-metrics',
                        description = "",
                        options = ('mae', 'mse', 'rmse', 'r2'))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass





import sys
import types
import numpy as np
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData

class ElementalPropertyAttributeGenerator:
    """Class to set up and generate descriptors based on elemental property
    statistics.

    Computes the mean, maximum, minimum, range, mode and mean
    absolute deviation of all elemental properties provided.

    Attributes
    ----------
    elemental_properties : array-like
        Elemental properties to be associated with this class for the
        generation of features.
    lookup-data : dict
        Dictionary containing the property name as the key and a list of
        floats as the value.

    """

    # Initialize to an empty list.
    elemental_properties = []

    def __init__(self, use_default_properties=True):
        """Class constructor.

        Parameters
        ----------
        use_default_properties : bool
            Flag to use default set of properties as defined below.

        """

        # Use default properties to reproduce Ward et al. descriptor values.
        if (use_default_properties):
            self.elemental_properties = ["Number", "MendeleevNumber",
                                         "AtomicWeight", "MeltingT",
                                         "Column", "Row", "CovalentRadius",
                                         "Electronegativity", "NsValence",
                                         "NpValence", "NdValence",
                                         "NfValence", "NValance",
                                         "NsUnfilled", "NpUnfilled",
                                         "NdUnfilled", "NfUnfilled",
                                         "NUnfilled", "GSvolume_pa",
                                         "GSbandgap", "GSmagmom",
                                         "SpaceGroupNumber"]

        # Initialize dictionary that will contain all the property values.
        self.lookup_data = {}

    def load_lookup_data(self):
        """Function to load the property values into `self.lookup_data` for the
        computation of features.

        """

        self.lookup_data = LookUpData.load_properties(
            self.elemental_properties)

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If no elemental properties are set.
            If input is not of type list.
            If items in the list are not CompositionEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Make sure that there is at least one elemental property provided.
        if not self.elemental_properties:
            raise ValueError("No elemental property is set. Add at least one "
                             "property to compute meaningful descriptors.")

        # If the dictionary containing the property values is empty,
        # load values into it.
        if not self.lookup_data:
            self.load_lookup_data()

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        # Insert header names here.
        n_statistics = 6
        for prop in self.elemental_properties:
            feat_headers.append("mean_"+prop)
            feat_headers.append("maxdiff_" + prop)
            feat_headers.append("dev_" + prop)
            feat_headers.append("max_" + prop)
            feat_headers.append("min_" + prop)
            feat_headers.append("most_" + prop)

        missing_data = {}
        # Generate features for each entry.
        for entry in entries:
            elem_ids = entry.get_element_ids()
            elem_fractions = entry.get_element_fractions()
            max_f = max(elem_fractions)
            tmp_list = []
            # Look up values for each property.
            for prop in self.elemental_properties:
                tmp_prop = []
                for elem_id in elem_ids:
                    tmp_prop_value = self.lookup_data[prop][elem_id]
                    # If data is missing, make a note of it so that we can
                    # inform the user later.
                    if np.isnan(tmp_prop_value):
                        if not prop in missing_data:
                            missing_data[prop] = []
                        if elem_id not in missing_data[prop]:
                            missing_data[prop].append(elem_id)
                    tmp_prop.append(tmp_prop_value)

                # If there is no missing data, compute statistics.
                if not np.isnan(tmp_prop).any():
                    mean_ = np.average(tmp_prop, weights=elem_fractions)
                    max_ = max(tmp_prop)
                    min_= min(tmp_prop)
                    max_diff_ = max_ - min_
                    avg_dev_ = np.average([abs(x - mean_) for x in tmp_prop],
                                          weights=elem_fractions)

                    indices = [i for i,f in enumerate(elem_fractions) if f >=
                               max_f]
                    most_ = sum(tmp_prop[i] for i in indices)/len(indices)

                    tmp_list.append(mean_)
                    tmp_list.append(max_diff_)
                    tmp_list.append(avg_dev_)
                    tmp_list.append(max_)
                    tmp_list.append(min_)
                    tmp_list.append(most_)
                else:
                    # Handle nan descriptors here from missing data.
                    for i in range(n_statistics):
                        tmp_list.append(np.nan)
            feat_values.append(tmp_list)

        # Issue warning to user about missing data here if it exists.
        if len(missing_data) > 0:
            sys.stderr.write("WARNING: There are " + str(len(missing_data)) +
                             " elemental properties with missing values: \n")
            for key in missing_data:
                sys.stderr.write("\t"+key+":")
                for elem in missing_data[key]:
                    sys.stderr.write(" "+LookUpData.element_names[elem])
                sys.stderr.write("\n")

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features

    def add_elemental_property(self, property):
        """Function to add an elemental property to `self.elemental_properties`
        in order to be used to compute features.

        Parameters
        ----------
        property : str
            Property to be added.

        """

        if property not in self.elemental_properties:
            self.elemental_properties.append(property)

    def add_elemental_properties(self, properties):
        """Function to provide a list of elemental properties to be used to
        compute features.

        Parameters
        ----------
        properties : array-like
            Properties to be included. A list of strings containing property
            names.

        """

        for prop in properties:
            self.add_elemental_property(prop)

    def remove_elemental_property(self, property):
        """
        Function to remove an elemental property from
        `self.elemental_properties`.

        Parameters
        ----------
        property : str
            Property to be removed.

        """

        if property in self.elemental_properties:
            self.elemental_properties.remove(property)

    def remove_elemental_properties(self, properties):
        """Function to remove a list of elemental properties from the list of
        elemental properties.

        Parameters
        ----------
        properties : array-like
            Properties to be removed. A list of strings containing property
            names.

        """

        for prop in properties:
            self.remove_elemental_property(prop)


import numpy as np

class AtomImage:
    """Class to uniquely identify an atom image. Key is atom ID, value is the
    periodic image.

    Attributes
    ----------
    position : array-like
        Pre-computed Cartesian coordinates of this image.
    atom : Atom
        Atom which this image is associated with.
    supercell : array-like
        Supercell in which this image is located.
    """

    def __init__(self, atom, image):
        """Constructor to create a new instance of the object.

        Parameters
        ----------
        atom : Atom
            Link to atom which is associated with this image.
        image : array-like
            Supercell position (i.e., which image it is in)
        """

        # Pre-computed Cartesian coordinates of this image.
        self.position = None

        # Atom which this image is associated with.
        self.atom = atom

        # Supercell in which this image is located.
        if not isinstance(image, np.ndarray):
            self.supercell = np.array(image, dtype=int).copy()
        else:
            self.supercell = image.copy()

        self.compute_position()

    def __cmp__(self, other):
        """Function that acts as the comparator for AtomImage objects.

        Parameters
        ----------
        other : AtomImage
            Other object.

        Returns
        -------
        output : int
            -1 is this < other, 0 if they are equal and +1 if this > other.
        """

        if self.atom.__eq__(other.atom):
            if self.supercell[0] != other.supercell[0]:
                return self.supercell[0] - other.supercell[0]
            elif self.supercell[1] != other.supercell[1]:
                return self.supercell[1] - other.supercell[1]
            else:
                return self.supercell[2] - other.supercell[2]
        else:
            return self.atom.get_id() - other.atom.get_id()

    def __eq__(self, other):
        """Function to override the check for equality with another object of
        the same class.

        Parameters
        ----------
        other : AtomImage
            Other object.

        Returns
        -------
        output : bool
            True if equal, else False.

        """

        if isinstance(other, AtomImage):
            return  self.atom.get_id == other.atom.get_id and np.array_equal(
                self.supercell, other.supercell)
        return False

    def __hash__(self):
        """Function to generate the hashcode for an object of this class.

        Returns
        -------
        output : int
            Hashcode value.
        """

        h = 7
        h = 43 * h + id(self.atom)
        for i in self.supercell:
            h = 31 * h + hash(i)
        return h

    def compute_position(self):
        """Function to compute (or re-compute) position of this image.

        """

        self.position = self.atom.get_cell().get_periodic_image(
            self.atom.get_position_cartesian(), self.supercell[0],
            self.supercell[1], self.supercell[2])

    def get_atom(self):
        """Function to get the atom at the center of this image.

        Returns
        -------
        output : atom
            Link to atom at the center.

        """

        return self.atom

    def get_atom_id(self):
        """Function to get the ID of the atom associated with this image.

        Returns
        -------
        output : int
            ID number.
        """

        return self.atom.get_id()

    def get_supercell(self):
        """Function to get which supercell this image is located in.

        Returns
        -------
        output : array-like
            Supercell coordinates.
        """

        return self.supercell.copy()

    def get_position(self):
        """Function to get the position (in Cartesian coordinates) of this image.

        Returns
        -------
        output : array-like
            Position of atom.
        """

        return self.position

    def __str__(self):
        """Function to override the output format for the builtin print command.

        Returns
        -------
        output : str
            Output string to be printed.
        """

        output = str(self.atom.get_id())
        image = self.get_supercell()

        if (image[0] == 0 and image[1] == 0 and image[2] == 0):
            return output

        output += "("
        for i in range(3):
            if image[i] != 0:
                output += "{0:d}{1:s}".format(image[i], str(chr(97+i)))
        output += ")"
        return output


from os.path import join, isfile
from os import listdir
from ...data.materials.CompositionEntry import CompositionEntry
import numpy as np
from ...data.materials.util.LookUpData import LookUpData
from ...vassal.analysis.VoronoiCellBasedAnalysis import \
    VoronoiCellBasedAnalysis
from ...vassal.io.VASP5IO import VASP5IO

class CrystalStructureEntry(CompositionEntry):
    """Class to represent a crystal structure.

    Attributes
    ----------
    structure : Cell
        Crystal structure represented in the form of a Cell.
    name : str
        Name given to denote this structure. Mainly used for book-keeping.
    radii : array-like
        List of radii (floats) for various atoms in the periodic table.
    voronoi : VoronoiCellBasedAnalysis
        Tool used to query properties of the tessellation.

    """

    # Crystal structure.
    structure = None

    # Name of entry.
    name = None

    # Link to atomic radii array.
    radii = None

    # Voronoi tessellation of this structure.
    voronoi = None

    def __init__(self, structure, name, radii):
        """Function to create an entry given its crystal structure.

        Parameters
        ----------
        structure : Cell
            Crystal structure represented in the form of a Cell.
        name : str
            Name given to denote this structure. Mainly used for book-keeping.
        radii : array-like
            List of radii (floats) for various atoms in the periodic table.
            Use None to leave radii unchanged.

        """
        super(CrystalStructureEntry, self).__init__()
        self.structure = structure
        self.name = name
        if structure.n_atoms() == 0:
            raise Exception("Cannot handle blank crystal structures.")
        self.radii = radii

        # Compute the composition of this crystal.
        self.compute_composition()

    def compute_composition(self):
        """Function to compute the composition of this crystal.

        Raises
        ------
        Exception
            If element name is not recognized.

        """

        # Get the composition.
        n_t = self.structure.n_types()
        elems = np.zeros(n_t, dtype=int)
        count = np.zeros(n_t, dtype=float)
        for i in range(n_t):
            name = self.structure.get_type_name(i)
            if name in LookUpData.element_ids:
                elems[i] = LookUpData.element_ids[name]
            else:
                raise Exception("Element name not recognized "+name)

            # Get the number of that atom.
            count[i] = float(self.structure.number_of_type(i))

            # Set the atomic radius of that atom.
            if self.radii is not None:
                self.structure.set_type_radius(i, self.radii[elems[i]])

        # Before reordering compositions.
        self.set_composition(count, element_ids=elems, to_sort=False)

    def replace_elements(self, replacements):
        """Function to create a new entry by replacing elements on this entry.

        Parameters
        ----------
        replacements : dict
            Dictionary of elements to replace. Key: Old element, Value: New
            element.

        Returns
        -------
        new_entry : CrystalStructureEntry
            New entry formed by replacing the current elements with the
            replacement map.

        """

        # Create new entry.
        new_entry = self.__copy__()
        new_entry.structure.replace_type_names(replacements)
        new_entry.structure.merge_like_types()
        new_entry.compute_composition()

        # If Voronoi tessellation has already been computed, create a tool
        # for the new entry w/o recomputing the tessellation.
        if self.voronoi is not None and (self.structure.n_types() !=
                                             new_entry.structure.n_types()):
            new_entry.voronoi = None
        return new_entry

    def __copy__(self):
        """Function to create a copy of this instance.

        Returns
        -------
        x : CrystalStructureEntry
            A copy of this instance.

        """

        x = type(self)(self.structure, self.name, self.radii)
        x.__dict__.update(self.__dict__)
        x.structure = self.structure.__copy__()
        return x

    def __eq__(self, other):
        """Function to check if this instance equals another instance.

        Parameters
        ----------
        other : CrystalStructureEntry
            Other composition entry to compare.

        Returns
        -------
        output : bool
            True if they are equal and False otherwise.

        """
        if isinstance(other, CrystalStructureEntry):
            return self.structure.__eq__(other.structure) and \
                   super(CrystalStructureEntry, self).__eq__(other)
        return False

    def __cmp__(self, other):
        """Function to compare this instance with another instance.

        Parameters
        ----------
        other : CrystalStructureEntry
            Other composition entry to compare.

        Returns
        -------
        output : int
            -1 if self < other , 1 if self > other or 0 if self = other.

        Raises
        ------
        Exception
            If the entries are equal.
        """

        if isinstance(other, CrystalStructureEntry):
            # First: Check for equality.
            if self.__eq__(other):
                return 0

            # Second: Check the composition / attributes.
            super_comp = super(CrystalStructureEntry, self).__cmp__(other)
            if super_comp != 0:
                return super_comp

            # Third: Extreme measures.
            if self.structure.n_atoms() != other.structure.n_atoms():
                return self.structure.n_atoms() - other.structure.n_atoms()
            if self.structure.n_types() != other.structure.n_types():
                return self.structure.n_types() - other.structure.n_types()
            for i in range(3):
                for j in range(3):
                    v1 = self.structure.get_basis()[i][j]
                    v2 = other.structure.get_basis()[i][j]
                    if (v1 - v2) ** 2 < 1e-30:
                        c = 0
                    elif v1 < v2:
                        c = -1
                    else:
                        c = +1

                    if c != 0:
                        return c

            for i in range(self.structure.n_atoms()):
                my_atom = self.structure.get_atom(i)
                your_atom = other.structure.get_atom(i)
                if my_atom.get_type() > your_atom.get_type():
                    return my_atom.get_type() - your_atom.get_type()
                my_pos = my_atom.get_position()
                your_pos = your_atom.get_position()
                for i in range(3):
                    v1 = my_pos[i]
                    v2 = your_pos[i]
                    if (v1 - v2) ** 2 < 1e-30:
                        c = 0
                    elif v1 < v2:
                        c = -1
                    else:
                        c = +1
            raise Exception("These entries were supposed to be unequal.")
        else:
            return super(CrystalStructureEntry, self).__cmp__(other)

    def __hash__(self):
        """Function to compute the hashcode of this instance.

        Returns
        -------
        output : int
            Hashcode of this instance.

        """
        return super(CrystalStructureEntry, self).__hash__() ^ \
               self.structure.__hash__()

    def get_structure(self):
        """Function to get link to the structure.

        Returns
        -------
        structure : Cell
            Structure this entry represents.
        """

        return self.structure

    def get_name(self):
        """Function to get the name of this entry.

        Returns
        -------
        name : str
            Name given to denote this structure. Mainly used for book-keeping.

        """
        return self.name

    def compute_voronoi_tessellation(self):
        """Function to compute the voronoi tessellation of this structure.

        Returns
        -------
        voronoi : VoronoiCellBasedAnalysis
            Tool used to query properties of the tessellation.

        """

        if self.voronoi is None:
            self.voronoi = VoronoiCellBasedAnalysis(radical=False)
            self.voronoi.analyze_structure(self.structure)
        elif not self.voronoi.tessellation_is_converged():
            raise Exception("Tessellation did not converge.")
        return self.voronoi

    def clear_representations(self):
        """Function to clear out the representations used when computing
        attributes.


        """

        self.voronoi = None

    def __str__(self):
        """Function the generate the string representation of this instance.

        Returns
        -------
        output : str
            Correctly formatted output.

        """

        comp = super(CrystalStructureEntry, self).__str__()
        return self.name+":"+comp

    @classmethod
    def import_structures_list(self, dir_path):
        """Function to read a list of crystal structures from a directory.

        Parameters
        ----------
        dir_path : str
            Path to the directory containing the list of vasp files.

        Returns
        -------
        structures_list : array-like
            A list of CrystalStructureEntry's corresponding to the file
            contents.

        """

        structures_list = []
        radii = LookUpData.load_property("CovalentRadius")

        # Thanks to https://stackoverflow.com/questions/3207219/how-do-i-list
        # -all-files-of-a-directory
        only_files = [f for f in listdir(dir_path) if isfile(join(dir_path,
                                                                  f))]
        io = VASP5IO()
        for f in only_files:
            if f.endswith(".vasp"):
                structure = io.parse_file(join(dir_path,f))
                name = f.split(".vasp")[0]
                entry = CrystalStructureEntry(structure, name=name, radii=radii)
                structures_list.append(entry)

        return structures_list